<?php

namespace Drupal\qtools_common\Plugin\Field\FieldFormatter;

use Drupal\Component\Utility\Unicode;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\qtools_common\Utility\HtmlTrimmer;

/**
 * Plugin implementation of the 'qtools_common_text_trimmed_ellipsis_formatter' formatter.
 *
 * Modified version of built-in formatter.
 *
 * @FieldFormatter(
 *   id = "qtools_common_text_trimmed_ellipsis_formatter",
 *   label = @Translation("Trimmed with ellipsis"),
 *   field_types = {
 *     "text",
 *     "text_long",
 *   },
 *   quickedit = {
 *     "editor" = "form"
 *   }
 * )
 */
class TextTrimmedEllipsisFormatter extends FormatterBase implements TrustedCallbackInterface {

  use StringTranslationTrait;

  const KEY_FORMATTER_SETTING = '#qtools_common_text_trimmed_ellipsis_formatter';

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'trim_length' => '80',
      'strip_html' => TRUE,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element['trim_length'] = [
      '#title' => $this->t('Trimmed limit'),
      '#type' => 'number',
      '#field_suffix' => $this->t('characters'),
      '#default_value' => $this->getSetting('trim_length'),
      '#description' => $this->t('If the summary is not set, the trimmed %label field will end at the last full sentence before this character limit.', ['%label' => $this->fieldDefinition->getLabel()]),
      '#min' => 1,
      '#required' => TRUE,
    ];
    $element['strip_html'] = [
      '#title' => $this->t('Strip HTML'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('strip_html'),
      '#description' => $this->t('Strip HTML before trimming.'),
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];
    $summary[] = $this->t('Trimmed limit: @trim_length characters, Strip HTML: @strip_html', [
      '@trim_length' => $this->getSetting('trim_length'),
      '@strip_html' => $this->getSetting('strip_html') ? $this->t('Yes') : $this->t('No'),
    ]);
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = [];

    $render_as_summary = function (&$element) {
      // Make sure any default #pre_render callbacks are set on the element,
      // because text_pre_render_summary() must run last.
      $element += \Drupal::service('element_info')->getInfo($element['#type']);
      // Add the #pre_render callback that renders the text into a summary.
      $element['#pre_render'][] = [static::class, 'preRenderTrimmed'];
      // Pass on the trim length to the #pre_render callback via a property.
      $element[static::KEY_FORMATTER_SETTING] = [
        'trim_length' => $this->getSetting('trim_length'),
        'strip_html' => $this->getSetting('strip_html'),
      ];
    };

    // The ProcessedText element already handles cache context & tag bubbling.
    // @see \Drupal\filter\Element\ProcessedText::preRenderText()
    foreach ($items as $delta => $item) {
      $elements[$delta] = [
        '#type' => 'processed_text',
        '#text' => NULL,
        '#format' => $item->format,
        '#langcode' => $item->getLangcode(),
      ];

      $elements[$delta]['#text'] = $item->value;
      $render_as_summary($elements[$delta]);
    }

    return $elements;
  }

  /**
   * Pre-render callback: Renders a processed text element's #markup.
   */
  public static function preRenderTrimmed(array $element) {
    $settings = $element[static::KEY_FORMATTER_SETTING];
    if (!empty($settings['strip_html'])) {
      $text = strip_tags($element['#markup']);
      $text = html_entity_decode($text, ENT_QUOTES);
      $text = Unicode::truncate($text, $settings['trim_length'], FALSE, TRUE);
      $element['#markup'] = htmlentities($text, ENT_QUOTES);
    }
    else {
      $element['#markup'] = HtmlTrimmer::trim($element['#markup'], $settings['trim_length']);
    }
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return ['preRenderTrimmed'];
  }

}
