<?php

namespace Drupal\qtools_common;

use Drupal\qtools_common\Utility\QToolsDebug;

/**
 * Class QToolsArrayHelper.
 *
 * @package Drupal\qtools
 */
abstract class QToolsArrayHelper {

  const MATCH_OP_EQUAL = '==';
  const MATCH_OP_EMPTY = 'empty';
  const MATCH_OP_ISSET = 'isset';
  const MATCH_OP_CONTAINS = 'contains';
  const MATCH_OP_PREG_MATCH = 'preg_match';
  const MATCH_OP_SIMILAR = 'similar';

  /**
   * Set array keys corresponding to one of item property/key.
   *
   * If element is missing key property/key it will be removed
   * from resulted array.
   *
   * @param array $array
   *   Array we want to remap.
   * @param string|array $key_path
   *   Keypath to new key value.
   * @param bool $group
   *   Specify if you want to group results with similar key into array.
   *
   * @return array
   *   Array keyed with new keys.
   */
  public static function remap(array $array, $key_path, $group = FALSE) {
    $res = [];

    // Return if given argument is not an array.
    if (!is_array($array)) {
      return $res;
    }

    // Remap array.
    foreach ($array as $value) {
      $new_key = static::getValue($value, $key_path);
      if (empty($group)) {
        $res[$new_key] = $value;
      }
      else {
        $res[$new_key][] = $value;
      }
    }
    return $res;
  }

  /**
   * Extract specified key from array.
   *
   * @param array $array
   *   Array of items, we will extract value from each of them.
   * @param string|array $key_path
   *   Path to value we search for (array of keys we need to check).
   * @param mixed $default
   *   Default value if we cant find given key_path in array.
   *
   * @return array
   *   Array of values for each item with given key_path.
   */
  public static function extract(array $array, $key_path, $default = NULL) {
    $res = [];

    // Return if given argument is not an array.
    if (!is_array($array)) {
      return $res;
    }

    // Extract given key.
    foreach ($array as $key => $value) {
      $res[$key] = static::getValue($value, $key_path, $default);
    }
    return $res;
  }

  /**
   * Get item from array(object) for specified path.
   *
   * @param array|object $array
   *   Item from which we will extract value.
   * @param string|array $key_path
   *   Path to value we search for (array of keys we need to check).
   * @param mixed $default
   *   Default value if we cant find given key_path in array.
   *
   * @return mixed
   *   Value of array with given key_path.
   */
  public static function getValue($array, $key_path, $default = NULL) {
    $array = (array) $array;

    $key_path = is_array($key_path) ? $key_path : [$key_path];
    $key = array_shift($key_path);

    $value = !empty($array[$key]) ? $array[$key] : $default;

    if (!empty($key_path) && (is_array($value) || is_object($value))) {
      $value = static::getValue($value, $key_path, $default);
    }
    return $value;
  }

  /**
   * Loads all elements in the given path.
   *
   * @return array
   *   References to loaded array elements.
   */
  public static function loadPath(&$array, $keys = []) {
    $scope = &$array;
    $parents[] = &$scope;
    foreach ($keys as $level) {
      $parents[] = &$scope[$level];
      $scope = &$scope[$level];
    }
    return $parents;
  }

  /**
   * Loads element for given path.
   *
   * @return array
   *   Loaded array elements.
   */
  public static function getNestedValue($arrayOrObject, $keys = [], $default = NULL) {
    $scope = $arrayOrObject;

    foreach ($keys as $level) {
      $scope = (array) $scope;
      if (array_key_exists($level, $scope)) {
        $scope = $scope[$level];
      }
      else {
        // If key does not exists return default value.
        return $default;
      }
    }

    // If we iterate over all keys with no error - return last scope.
    return $scope;
  }

  /**
   * Create or modify match rules.
   *
   * @return array
   *   Match rules.
   */
  public static function matchRules($key, $value, $op = self::MATCH_OP_EQUAL, $rules = []) {
    $rules[] = [$key, $op, $value];
    return $rules;
  }

  /**
   * Searches array for matched elements.
   *
   * @return array
   *   Result with links to found elements and their respective paths.
   */
  public static function search($array, $filters, $lister = 'array_keys') {
    // Call inner filter function.
    $path = [];
    $found = [];
    static::searchInner($array, $filters, $lister, $path, $found);
    return $found;
  }

  /**
   * Inner search function.
   *
   * @return array|null
   *   Result with links to found elements and their respective paths.
   */
  protected static function searchInner(&$array, $filters, $lister, &$path, &$found) {
    // Exit if element is not an array.
    if (!is_array($array)) {
      return NULL;
    }

    // Check if current element match filters.
    if (static::elementMatchFilters($array, $filters)) {
      $found[] = [
        'path' => $path,
        'element' => &$array,
      ];
    }

    // Check child elements.
    foreach ($lister($array) as $key) {
      if (is_array($array[$key])) {
        $sub_path = $path;
        $sub_path[] = $key;
        static::searchInner($array[$key], $filters, $lister, $sub_path, $found);
      }
    }
  }

  /**
   * Check if element matches set of filters.
   *
   * @return bool
   *   Result of the match operation.
   */
  protected static function elementMatchFilters(&$element, $filters, $and = TRUE, $invert = FALSE) {
    $match = $and ? TRUE : FALSE;

    foreach ($filters as $filter) {
      $match = $and ?
        $match && static::elementMatchFilterInner($element, $filter) :
        $match || static::elementMatchFilterInner($element, $filter);
    }

    return $invert ? !$match : $match;
  }

  /**
   * Check if element matches one filter.
   *
   * @return bool
   *   Result of the match operation.
   */
  protected static function elementMatchFilterInner(&$element, $filter) {
    $match = FALSE;

    $key = isset($filter[0]) ? $filter[0] : NULL;
    $op = isset($filter[1]) ? $filter[1] : NULL;
    $value = isset($filter[2]) ? $filter[2] : NULL;
    $invert = isset($filter[3]) ? $filter[3] : FALSE;
    switch ($op) {
      case static::MATCH_OP_ISSET:
        $match = isset($element[$key]);
        break;

      case static::MATCH_OP_EMPTY:
        $match = empty($element[$key]);
        break;

      case static::MATCH_OP_CONTAINS:
        $match = isset($element[$key]) && (strpos($element[$key], $value) !== FALSE);
        break;

      case static::MATCH_OP_PREG_MATCH:
        if (isset($element[$key])) {
          $match = (bool) preg_match($value, $element[$key]);
        }
        break;

      case static::MATCH_OP_SIMILAR:
        if (isset($element[$key])) {
          $input = strtolower($value);
          $compare = strtolower($element[$key]);
          $match = (strpos($compare, $input) !== FALSE);
        }
        break;

      default:
        $match = isset($element[$key]) && ($element[$key] === $value);
        break;
    }

    return $invert ? !$match : $match;
  }

  /**
   * Split array into chunks, using values array as separators.
   *
   * @return array
   *   Chunks array.
   */
  public static function splitByValues(array $array, array $split_values) {
    $result = [];
    $index = 0;
    foreach ($array as $key => $value) {
      if (in_array($value, $split_values)) {
        $index++;
      }
      else {
        $result[$index][$key] = $value;
      }
    }
    return $result;
  }

}
