<?php

namespace Drupal\qtools_common;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\Context\CacheContextsManager;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\Core\Entity\Entity;
use Drupal\Core\Entity\EntityInterface;

/**
 * Class QToolsFormHelper.
 *
 * @package Drupal\qtools
 */
abstract class QToolsCacheHelper {

  const CACHE_FOUND_NONE = 0;
  const CACHE_FOUND_STATIC = 1;
  const CACHE_FOUND_PERSISTENT = 2;

  // One and two seconds cache unavailable but who cares?
  const TTL_NONE = 0;
  const TTL_STATIC = 1;
  const TTL_PERMANENT = 2;

  const MAX_AGE_NO_CACHE = 0;

  const STATIC_POOL_DEFAULT = 'Drupal\qtools_common\QToolsCacheHelper';
  const PERSISTENT_PREFIX = '_';

  // CID prefix for cache clear check.
  const CID_PREFIX_TAGS_CLEAR_TRACK = 'qtools_common_tag_clear';

  /**
   * Sets cache with metadata.
   */
  public static function setWithMetadata(array $cid_parts, $data, CacheableMetadata $cacheableMetadata, $cache_bin = 'default') {
    // Parse metadata.
    $contexts = $cacheableMetadata->getCacheContexts();
    $max_age = $cacheableMetadata->getCacheMaxAge();
    $tags = $cacheableMetadata->getCacheTags();

    // Get previous contexts.
    $contexts_cid = static::buildCid($cid_parts);
    $cached_contexts = \Drupal::cache($cache_bin)->get($contexts_cid);
    if (!empty($cached_contexts)) {
      $contexts = Cache::mergeContexts($contexts, $cached_contexts->data);
    }

    // Save contexts (this is not affected by neither max-age nor tags).
    \Drupal::cache($cache_bin)->set($contexts_cid, $contexts, Cache::PERMANENT);

    // Save data content.
    $data_cid = static::buildCid($cid_parts, $contexts);
    \Drupal::cache($cache_bin)->set($data_cid, $data, static::maxAgeToExpire($max_age), $tags);
  }

  /**
   * Gets cache with metadata.
   *
   * @return false|object
   *   Cached object or false.
   */
  public static function getWithMetadata(array $cid_parts, CacheableMetadata $cacheableMetadata = NULL, $allow_invalid = FALSE, $cache_bin = 'default') {
    $cached_data = FALSE;

    // Load metadata.
    $contexts_cid = QToolsCacheHelper::buildCid($cid_parts);
    $cached_contexts = \Drupal::cache($cache_bin)->get($contexts_cid);

    // Get layer data.
    if (!empty($cached_contexts)) {
      $contexts = $cached_contexts->data;
      $cacheableMetadata->addCacheContexts($contexts);

      // Load content.
      $data_cid = QToolsCacheHelper::buildCid($cid_parts, $contexts);
      $cached_data = \Drupal::cache($cache_bin)->get($data_cid, $allow_invalid);
      if (!empty($cached_data) && !empty($cacheableMetadata)) {
        // Rebuild valid cacheable metadata.
        $cacheableMetadata->addCacheTags($cached_data->tags);
        $cacheableMetadata->setCacheMaxAge(static::expireToMaxAge($cached_data->expire));
      }
    }

    return $cached_data;
  }

  /**
   * Extends metadata with values from another metadata.
   */
  public static function extendMetadata(RefinableCacheableDependencyInterface $cacheableMetadata, CacheableDependencyInterface $additionalMetadataObject = NULL) {
    if (!empty($additionalMetadataObject)) {
      $cacheableMetadata->addCacheTags($additionalMetadataObject->getCacheTags());
      $cacheableMetadata->addCacheContexts($additionalMetadataObject->getCacheContexts());
      $cacheableMetadata->mergeCacheMaxAge($additionalMetadataObject->getCacheMaxAge());
    }
  }

  /**
   * Shrinks metadata from some contexts and tags.
   */
  public static function shrinkMetadata(CacheableMetadata $cacheableMetadata, $excluded_contexts = [], $excluded_tags = []) {
    if (!empty($excluded_contexts)) {
      $contexts = $cacheableMetadata->getCacheContexts();
      $contexts = array_diff($contexts, $excluded_contexts);
      $cacheableMetadata->setCacheContexts($contexts);
    }
    if (!empty($excluded_tags)) {
      $tags = $cacheableMetadata->getCacheTags();
      $tags = array_diff($tags, $excluded_tags);
      $cacheableMetadata->setCacheTags($tags);
    }
  }

  /**
   * Extend with entity listings tags.
   */
  public static function extendWithEntityListingTags(RefinableCacheableDependencyInterface $metadata, array $bundles, string $entity_type = 'node') {
    $tags = [];
    foreach ($bundles as $bundle) {
      $tags[] = $entity_type . '_list:' . $bundle;
    }
    $metadata->addCacheTags($tags);
  }

  /**
   * Build bundle specific listing tags for entity.
   *
   * @see https://www.drupal.org/node/2145751
   *
   * @return array
   *   Array of bundle specific listing tags.
   */
  public static function getListBundleCacheTags(EntityInterface $entity = NULL, $entity_type = NULL, $bundle = NULL) {
    // If entity is passed we take info from it.
    if (!empty($entity) && ($entity->bundle() != $entity->getEntityTypeId())) {
      $entity_type = $entity->getEntityTypeId();
      $bundle = $entity->bundle();
    }

    // Build custom tags for bundle specific list.
    if (!empty($entity_type) && !empty($bundle)) {
      $tags = [$entity_type . '_list:' . $bundle];
    }
    else {
      $tags = [];
    }

    return $tags;
  }

  /**
   * Load form static cache.
   *
   * @return array
   *   Returns data and cache operation.
   */
  protected static function staticLoad(array $info, CacheableMetadata $cacheableMetadata) {
    // Access static cache.
    $static_cache = &drupal_static($info['static_pool'], []);

    // Load from cache.
    if (isset($static_cache[$info['static_cid']])) {
      $data = $static_cache[$info['static_cid']]['data'];
      QToolsCacheHelper::extendMetadata($cacheableMetadata, $static_cache[$info['static_cid']]['cacheableMetadata']);
      $op = static::CACHE_FOUND_STATIC;
    }
    else {
      $data = NULL;
      $op = static::CACHE_FOUND_NONE;
    }

    return [$data, $op];
  }

  /**
   * Clears record from both static and persistent cache.
   */
  public static function cachedCallClear($cid_keys, $static_pool = self::STATIC_POOL_DEFAULT, $persistent_preffix = self::PERSISTENT_PREFIX, $cache_bin = 'default') {
    // Clear static pool.
    $cids = static::cachedCallCid($cid_keys, $static_pool, $persistent_preffix);
    $static_cache = &drupal_static($static_pool, []);

    if (isset($static_cache[$cids['static_cid']])) {
      unset($static_cache[$cids['static_cid']]);
    }

    // Clear persistent cache.
    \Drupal::cache($cache_bin)->delete($cids['cid']);
  }

  /**
   * Clears static pool.
   *
   * @param string $pool
   *   Name of static pool, usually ($class . '.' . $method) name.
   */
  public static function staticClear($pool) {
    $static_cache = &drupal_static($pool, []);
    $static_cache = [];
  }

  /**
   * Updates static cache.
   */
  protected static function staticUpdate($info, $data, CacheableMetadata $cacheableMetadata) {
    // Access static cache.
    $static_cache = &drupal_static($info['static_pool'], []);
    $static_cache[$info['static_cid']] = [
      'data' => $data,
      'cacheableMetadata' => $cacheableMetadata,
    ];
  }

  /**
   * Load form persistent cache.
   *
   * @return array
   *   Array that contains $data and $op cache operation.
   */
  protected static function persistentLoad($info, CacheableMetadata $cacheableMetadata) {
    // Check if we have this object in normal drupal cache.
    if ($cache = \Drupal::cache($info['cache_bin'])->get($info['cid'])) {
      $data = $cache->data['data'];
      $tags = $cache->data['tags'];
      $expire = $cache->data['expire'];
      $contexts = $cache->data['contexts'];

      // Subtract current time from expiration to get max age.
      $max_age = ($expire === Cache::PERMANENT) ?
        Cache::PERMANENT :
        $expire - (int) \Drupal::requestStack()->getMasterRequest()->server->get('REQUEST_TIME');

      $cacheableMetadata->addCacheTags($tags);
      $cacheableMetadata->addCacheContexts($contexts);
      $cacheableMetadata->mergeCacheMaxAge($max_age);

      $op = static::CACHE_FOUND_PERSISTENT;
    }
    else {
      $data = NULL;
      $op = static::CACHE_FOUND_NONE;
    }
    return [$data, $op];
  }

  /**
   * Convert max age to expire.
   *
   * @return int
   *   Expire timestamp.
   */
  public static function maxAgeToExpire($max_age) {
    if ($max_age == static::MAX_AGE_NO_CACHE || $max_age == Cache::PERMANENT) {
      return $max_age;
    }
    return (int) \Drupal::requestStack()->getMasterRequest()->server->get('REQUEST_TIME') + $max_age;
  }

  /**
   * Convert expire to max age.
   *
   * @return int
   *   max-age.
   */
  public static function expireToMaxAge($expire) {
    if ($expire == Cache::PERMANENT) {
      return $expire;
    }
    return (int) $expire - \Drupal::requestStack()->getMasterRequest()->server->get('REQUEST_TIME');
  }

  /**
   * Update persistent cache.
   */
  protected static function persistentUpdate($info, $data, CacheableMetadata $cacheableMetadata) {
    // Save to persistent cache.
    $expire = static::maxAgeToExpire($cacheableMetadata->getCacheMaxAge());

    // Skip cache if expire empty.
    if (empty($expire)) {
      return;
    }

    // Store cache.
    $tags = $cacheableMetadata->getCacheTags();
    $contexts = $cacheableMetadata->getCacheContexts();

    \Drupal::cache($info['cache_bin'])->set($info['cid'], [
      'data' => $data,
      'contexts' => $contexts,
      'expire' => $expire,
      'tags' => $tags,
    ], $expire, $tags);
  }

  /**
   * Run normal data calculation.
   *
   * @return array
   *   Array that contains $data, $cache_tags and optionally $new_ttl
   */
  protected static function normalLoad($info, CacheableMetadata $cacheableMetadata) {
    // Pass metadata inside target function as last parameter.
    $info['params'][] = $cacheableMetadata;

    if (empty($info['class'])) {
      // Simple functions.
      $result = $info['callable'](...$info['params']);
    }
    elseif ($info['class'] == 'Closure') {
      // Anonymous functions.
      $result = $info['callable'][0](...$info['params']);
    }
    else {
      // Object+Method.
      $result = call_user_func_array($info['callable'], $info['params']);
    }

    return $result;
  }

  /**
   * Convert QTools TTL into Drupals max-age.
   */
  protected static function ttlToMaxAge($ttl) {
    if ($ttl == static::TTL_PERMANENT) {
      return Cache::PERMANENT;
    }
    elseif ($ttl == static::TTL_STATIC) {
      return static::MAX_AGE_NO_CACHE;
    }
    return $ttl;
  }

  /**
   * Add cacheable metadata to build without destroying existing metadata in it.
   */
  public static function addCacheableMetadataToBuild(CacheableMetadata $cacheableMetadata, array &$build): array {
    $buildMetadata = CacheableMetadata::createFromRenderArray($build);
    $buildMetadata->addCacheableDependency($cacheableMetadata);
    $buildMetadata->applyTo($build);
    return $build;
  }

  /**
   * Builds cache CID respecting current contexts.
   *
   * @return string
   *   Cache cid with context resolved into values.
   */
  public static function buildCid(array $cid_parts, array $contexts = NULL) {
    if (empty($cid_parts)) {
      return NULL;
    }

    // Add contexts values if specified.
    if (!empty($contexts)) {
      $context_parts = static::convertContexts($contexts);
      $cid_parts = array_merge($cid_parts, $context_parts);
    }

    return implode(':', $cid_parts);
  }

  /**
   * Convert contexts to values.
   *
   * @return array
   *   Cache contexts values.
   */
  public static function convertContexts(array $contexts = NULL) {
    if (empty($contexts)) {
      return [];
    }

    /** @var \Drupal\Core\Cache\Context\CacheContextsManager $cacheContextsManager */
    $cacheContextsManager = \Drupal::service('cache_contexts_manager');
    $context_cache_keys = $cacheContextsManager->convertTokensToKeys($contexts);
    return $context_cache_keys->getKeys();
  }

  /**
   * Builds cache CID respecting current contexts.
   *
   * @return string
   *   Cache cid with context resolved into values.
   */
  public static function buildCidElements($elements, $cid_parts = []) {
    if (empty($cid_parts)) {
      $cid_parts = !empty($elements['#cache']['keys']) ? $elements['#cache']['keys'] : [];
    }
    $contexts = !empty($elements['#cache']['contexts']) ? $elements['#cache']['contexts'] : [];

    return static::buildCid($cid_parts, $contexts);
  }

  /**
   * Builds cached call cid.
   */
  protected static function cachedCallCid($cid_keys, $static_pool, $persistent_preffix) {
    $cid['static_cid'] = json_encode($cid_keys);
    $persistent_suffix = strlen($cid['static_cid']) > 30 ?
      md5($cid['static_cid']) :
      str_replace(':', '-', $cid['static_cid']);
    $cid['cid'] = $static_pool . ':' . $persistent_preffix . ':' . $persistent_suffix;

    return $cid;
  }

  /**
   * Issue cached call.
   *
   * @return array
   *   Array that contains $data and $op cache operation.
   */
  public static function cachedCall($callable, $params, $cid_keys, $ttl = self::TTL_STATIC, $static_pool = self::STATIC_POOL_DEFAULT, $persistent_preffix = self::PERSISTENT_PREFIX, $cache_bin = 'default') {
    $op = static::CACHE_FOUND_NONE;
    $data = NULL;
    $cacheableMetadata = new CacheableMetadata();

    $cids = static::cachedCallCid($cid_keys, $static_pool, $persistent_preffix);

    // Build cache info as late as possible.
    $info = [
      'callable' => $callable,
      'params' => $params,
      'cid_keys' => $cid_keys,
      'cache_bin' => $cache_bin,
    ];
    if (is_array($callable)) {
      $info['class'] = get_class($callable[0]);
      $info['method'] = $callable[1];
    }
    else {
      $info['class'] = '';
      $info['method'] = $callable;
    }

    // Check for static cache.
    if ($ttl >= static::TTL_STATIC) {
      $info['static_pool'] = $static_pool;
      $info['static_cid'] = $cids['static_cid'];
      [$data, $op] = static::staticLoad($info, $cacheableMetadata);

      // If no static cache, and TTL ALLOWS for more than static.
      if ($ttl >= static::TTL_PERMANENT && $op == static::CACHE_FOUND_NONE) {
        $info['cid'] = $cids['cid'];
        [$data, $op] = static::persistentLoad($info, $cacheableMetadata);
      }
    }

    // If cache was still not loaded (operation is not filled),
    // we load it normally.
    if ($op == static::CACHE_FOUND_NONE) {
      // Set metadata max_age based on TTL.
      $cacheableMetadata->mergeCacheMaxAge(static::ttlToMaxAge($ttl));
      $data = static::normalLoad($info, $cacheableMetadata);

      // Update persistent cache if we should cache for more than static.
      if ($ttl >= static::TTL_PERMANENT) {
        static::persistentUpdate($info, $data, $cacheableMetadata);
      }
    }

    // Update static cache if we should cache.
    // Update from CACHE_FOUND_NONE and CACHE_FOUND_PERSISTENT.
    if ($ttl >= static::TTL_STATIC && $op != static::CACHE_FOUND_STATIC) {
      static::staticUpdate($info, $data, $cacheableMetadata);
    }

    return [$data, $op, $cacheableMetadata, $info];
  }

  /**
   * Build cacheable response object.
   *
   * @return array
   *   Array with specific keys suitable for cacheable response
   *   (static::normalLoad) parser.
   */
  public static function buildCacheableResponse($data, $tags, $new_ttl = NULL) {
    return [
      '##data' => $data,
      '##tags' => $tags,
      '##ttl' => $new_ttl,
    ];
  }

  /**
   * Checks when cache update was triggered for specific tags.
   *
   * Note that is will return timestamp always greater or equal
   * of cache clear action.
   *
   * @return int
   *   Time that is greater or equal of cache clear for one of asked tags.
   */
  public static function getCacheTagsUpdatedUpperBound($tags, $now = NULL, $cache_bin = 'default') {
    sort($tags);
    $cid = implode('_', array_merge([static::CID_PREFIX_TAGS_CLEAR_TRACK], $tags));

    // Get last tag clear.
    $data = \Drupal::cache($cache_bin)->get($cid);
    if ($data) {
      return $data->data;
    }

    // If cache was cleared after last timestamp set, $data above will be empty,
    // use $now or time() to signal that cache was cleared somewhere before now.
    $now = $now ?: time();
    \Drupal::cache($cache_bin)->set($cid, $now, Cache::PERMANENT, $tags);
    return $now;
  }

}
