<?php

namespace Drupal\qtools_common;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Site\Settings;
use Drupal\qtools_common\Utility\QToolsConfigArrayMapType;
use Drupal\qtools_common\Utility\Slug;

/**
 * Class QToolsConfigHelper.
 *
 * @package Drupal\qtools
 */
abstract class QToolsConfigHelper {
  const SOURCE_SESSION = 'session';
  const SOURCE_STATE = 'state';
  const SOURCE_SETTINGS = 'settings';
  const SOURCE_CONFIG_OVERRIDES = 'config_overrides';
  const SOURCE_CONFIG = '';
  const SOURCE_DEFAULT = 'default';

  const HOOK_QTOOLS_CONFIG_HELPER_DEFINITIONS = 'qtools_common_config_definitions';

  const NAME = '__NAME__';

  const STATE_ACTIVE = 1;
  const STATE_OVERRIDEN = -1;
  const STATE_NOT_SET = 0;

  const NO_COLLISION_PREFIX = 'qtools_config_helper_overrides';

  const KEY_SUFFIX_GROUP = '_group';
  const KEY_SUFFIX_OVERRIDE_ENABLED = '_override_enabled';

  const TYPE_ARRAY_LIST = 'array_list';
  const TYPE_ARRAY_MAP = 'array_map';

  const ARRAY_MAP_KEY_VALUE_SEPARATOR = '|';

  /**
   * Packs simple array int string representation.
   *
   * @param array $array
   *   Input array.
   * @param bool $clean
   *   If true, trim incoming values.
   *
   * @return string
   *   Packed array.
   */
  protected static function packArrayList(array $array = [], $clean = TRUE) {
    if ($clean) {
      foreach ($array as $key => $value) {
        $array[$key] = trim($value);
      }
    }
    return implode(PHP_EOL, $array);
  }

  /**
   * Unpacks simple packed array.
   *
   * @param string $input
   *   Packed array from packArrayList() function.
   * @param bool $clean
   *   If true, will apply trim and remove empty values from result.
   *
   * @return array
   *   Unpacked array
   */
  protected static function unpackArrayList($input = '', $clean = TRUE) {
    $unpacked = explode(PHP_EOL, $input);
    if ($clean) {
      foreach ($unpacked as $key => $value) {
        $unpacked[$key] = trim($value);
      }
      $unpacked = array_filter($unpacked);
    }
    return $unpacked;
  }

  /**
   * Packs map array int string representation.
   *
   * @param array $array
   *   Input array.
   * @param bool $clean
   *   If true, trim incoming values.
   *
   * @return string
   *   Packed array.
   */
  protected static function packArrayMap(array $array = [], $clean = TRUE) {
    $packed = [];
    foreach ($array as $key => $value) {
      $packed[] = $key . static::ARRAY_MAP_KEY_VALUE_SEPARATOR . ($clean ? trim($value) : $value);
    }
    return implode(PHP_EOL, $packed);
  }

  /**
   * Unpacks simple packed array.
   *
   * @param string $input
   *   Packed array from packArrayMap() function.
   * @param bool $clean
   *   If true, will apply trim and remove empty values from result.
   *
   * @return array
   *   Unpacked array
   */
  protected static function unpackArrayMap($input = '', $clean = TRUE) {
    $lines = explode(PHP_EOL, $input);
    $lines = array_filter($lines);
    $unpacked = [];

    foreach ($lines as $line) {
      $parts = explode(static::ARRAY_MAP_KEY_VALUE_SEPARATOR, $line);
      $key = array_shift($parts);
      $value = implode(static::ARRAY_MAP_KEY_VALUE_SEPARATOR, $parts);
      $unpacked[$key] = $clean ? trim($value) : $value;
    }

    return $unpacked;
  }

  /**
   * Creates default config definitions.
   *
   * @param string $config_name
   *   Configuration name/key.
   * @param mixed $global_default
   *   Global Default for given config.
   * @param array $extra
   *   Additional params.
   *
   * @return array
   *   Default config definition.
   */
  public static function defaultDefinition($config_name, $global_default, array $extra = []) {
    [$name, $key, $parents] = static::parseName($config_name);

    if ($global_default instanceof QToolsConfigArrayMapType) {
      $type = static::TYPE_ARRAY_MAP;
    }
    elseif (is_array($global_default)) {
      $type = static::TYPE_ARRAY_LIST;
    }
    elseif (strpos($global_default, PHP_EOL) !== FALSE) {
      $type = 'text';
      $global_default = trim($global_default);
    }
    else {
      $type = gettype($global_default);
    }
    return [
      static::NAME => $name,
      'key' => $key,
      'sources' => [
        [static::SOURCE_SESSION, $name, NULL],
        [static::SOURCE_STATE, $name, NULL],
        [static::SOURCE_SETTINGS, $name, NULL],
        [static::SOURCE_CONFIG_OVERRIDES, $name, NULL],
        [static::SOURCE_CONFIG, $name, NULL],
        [static::SOURCE_DEFAULT, $global_default, NULL],
      ],
      'type' => $type,
      'extra' => $extra,
      'parents' => $parents,
    ];
  }

  /**
   * Parse config name.
   *
   * @param string $config_name
   *   Configuration name/key.
   *
   * @return array
   *   Parsed [$name, $key, $parents].
   */
  protected static function parseName($config_name) {
    $name = str_replace("'", '', $config_name);
    if (strpos($name, '][') !== FALSE) {
      $parents = explode('][', $name);
      $key = end($parents);
    }
    else {
      $parents = [$name];
      $key = $name;
    }
    return [$name, $key, $parents];
  }

  /**
   * Gather all available configs.
   *
   * @return array
   *   List of all available configs.
   */
  protected static function getConfigsList() {
    $config_list = &drupal_static(__CLASS__ . __METHOD__, NULL);
    if (!empty($config_list)) {
      return $config_list;
    }

    /** @var \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler */
    $moduleHandler = \Drupal::service('module_handler');
    $all = $moduleHandler->invokeAll(static::HOOK_QTOOLS_CONFIG_HELPER_DEFINITIONS);

    $tree = [];
    $list = [];
    foreach ($all as $group => $definitions) {
      foreach ($definitions as $definition) {
        $definition['group'] = $group;
        NestedArray::setValue($tree, $definition['parents'], $definition);
        $list[$definition[static::NAME]] = $definition;
      }
    }

    $config_list = [$list, $tree];
    return $config_list;
  }

  /**
   * Make sure that config type will be predictable.
   *
   * @param mixed $value
   *   Current config state value.
   * @param array $definition
   *   Configuration definition.
   *
   * @return bool|float|int|string
   *   Config with binded type.
   */
  protected static function bindValueType($value, array $definition) {
    switch ($definition['type']) {
      case 'double':
        $value = (float) $value;
        break;

      case 'integer':
        $value = (int) $value;
        break;

      case 'boolean':
        $value = (bool) $value;
        break;

      case static::TYPE_ARRAY_MAP:
        $value = (array) $value;
        break;

      case static::TYPE_ARRAY_LIST:
        $value = (array) $value;
        break;

      // Strings and any unknown ones.
      default:
        $value = (string) $value;
        break;

    }

    return $value;
  }

  /**
   * Get current configuration states.
   *
   * @param array $definition
   *   Config definition.
   *
   * @return array
   *   All configuration states.
   */
  public static function getConfigStates(array $definition) {
    $states = [];
    foreach ($definition['sources'] as $source) {
      [$type, $source_configuration, $null_value] = $source;

      if ($type == static::SOURCE_DEFAULT) {
        $value = $source_configuration;
      }
      else {
        [$name, $key, $parents] = static::parseName($source_configuration);

        $value = $null_value;
        switch ($type) {
          case static::SOURCE_SESSION:
            $value = static::populateValueFromSession($name, $key, $parents);
            break;

          case static::SOURCE_STATE:
            $value = static::populateValueFromState($name, $key, $parents);
            break;

          case static::SOURCE_SETTINGS:
            $value = static::populateValueFromSettings($name, $key, $parents);
            break;

          case static::SOURCE_CONFIG_OVERRIDES:
            $value = static::populateValueFromConfigurationOverrides($name, $key, $parents);
            break;

          case static::SOURCE_CONFIG:
            $value = static::populateValueFromConfiguration($name, $key, $parents);
            break;

          case static::SOURCE_DEFAULT:
            $value = static::populateValueFromConfiguration($name, $key, $parents);
            break;

          default:
            // Do nothing, fire warning.
            break;
        }
      }
      // Apply value type binding.
      if ($value !== NULL) {
        $value = static::bindValueType($value, $definition);
      }

      if ($value === $null_value) {
        $value = NULL;
      }
      $states[$type]['value'] = $value;
    }

    // Find active one.
    $active_state = NULL;
    $fallback_state = NULL;
    foreach ($states as $type => $state) {
      // Get active state.
      if (empty($active_state) && $state['value'] !== NULL) {
        $active_state = $type;
        $states[$type]['status'] = static::STATE_ACTIVE;
      }
      elseif ($state['value'] !== NULL) {
        $states[$type]['status'] = static::STATE_OVERRIDEN;
      }
      else {
        $states[$type]['status'] = static::STATE_NOT_SET;
      }
    }

    return [$states, $active_state];
  }

  /**
   * Populates confid with values from sources.
   *
   * @param array $definition
   *   Configuration definition.
   * @param mixed $local_default
   *   Local default value.
   *
   * @return mixed
   *   Active configuration value.
   */
  protected static function populateValue(array $definition, $local_default = NULL) {
    $cache = &drupal_static('qtools_common.config_helper.configs', []);
    $cid = $definition[static::NAME];
    if (!empty($cache[$cid])) {
      return $cache[$cid];
    }

    [$states, $active_state] = static::getConfigStates($definition);

    $cache[$cid] = ($active_state == static::SOURCE_DEFAULT) ?
      ($local_default ?? $states[$active_state]['value']) :
      $states[$active_state]['value'];
    return $cache[$cid];
  }

  /**
   * Populate config value from state.
   *
   * @param string $name
   *   Config name.
   * @param string $key
   *   Config key.
   * @param array $parents
   *   Config parents in definition tree.
   *
   * @return mixed
   *   Config value.
   */
  protected static function populateValueFromState($name, $key, array $parents) {
    return \Drupal::state()->get(static::getNoCollisionName($name)) ?? \Drupal::state()->get($key);
  }

  /**
   * Populate config value from session.
   *
   * @param string $name
   *   Config name.
   * @param string $key
   *   Config key.
   * @param array $parents
   *   Config parents in definition tree.
   *
   * @return mixed
   *   Config value.
   */
  protected static function populateValueFromSession($name, $key, array $parents) {
    return $_SESSION['qtools_common'][static::NO_COLLISION_PREFIX][$name] ?? NULL;
  }

  /**
   * Populate config value from settings.
   *
   * @param string $name
   *   Config name.
   * @param string $key
   *   Config key.
   * @param array $parents
   *   Config parents in definition tree.
   *
   * @return mixed
   *   Config value.
   */
  protected static function populateValueFromSettings($name, $key, array $parents) {
    $settings = Settings::getAll();
    return NestedArray::getValue($settings, $parents);
  }

  /**
   * Populate config value from overridden configuration.
   *
   * @param string $name
   *   Config name.
   * @param string $key
   *   Config key.
   * @param array $parents
   *   Config parents in definition tree.
   *
   * @return mixed
   *   Config value.
   */
  protected static function populateValueFromConfigurationOverrides($name, $key, array $parents) {
    $value = NULL;

    // Configs require at least two level nesting.
    if (count($parents) >= 2) {
      // See ConfigFactory.php.
      $overrides = $GLOBALS['config'][$parents[0]] ?? [];
      $bucket = $overrides[$parents[1]] ?? NULL;
      if (count($parents) > 2) {
        $tail = array_slice($parents, 2);
        $value = NestedArray::getValue($bucket, $tail);
      }
      else {
        $value = $bucket;
      }
    }
    return $value;
  }

  /**
   * Populate config value from configuration.
   *
   * @param string $name
   *   Config name.
   * @param string $key
   *   Config key.
   * @param array $parents
   *   Config parents in definition tree.
   *
   * @return mixed
   *   Config value.
   */
  protected static function populateValueFromConfiguration($name, $key, array $parents) {
    $value = NULL;

    // Configs require at least two level nesting.
    if (count($parents) >= 2) {
      $bucket = \Drupal::config($parents[0])->get($parents[1]);
      if (count($parents) > 2) {
        $tail = array_slice($parents, 2);
        $value = NestedArray::getValue($bucket, $tail);
      }
      else {
        $value = $bucket;
      }
    }
    return $value;
  }

  /**
   * Populate config values including subtree.
   *
   * @param array $definition
   *   Config definition.
   * @param mixed $local_default
   *   Local default.
   *
   * @return mixed
   *   Config value.
   */
  protected static function populateValues(array $definition, $local_default = NULL) {
    $nested = [];
    if (isset($definition[static::NAME])) {
      // Local default is only used once if we getting value directly.
      return static::populateValue($definition, $local_default);
    }
    else {
      foreach ($definition as $key => $sub_def) {
        $nested[$key] = self::populateValues($sub_def);
      }
    }
    return $nested;
  }

  /**
   * Get configuration value.
   *
   * @param string $config_name
   *   Config name.
   * @param mixed $local_default
   *   Local default value.
   *
   * @return mixed|null
   *   Config value.
   */
  public static function get($config_name, $local_default = NULL) {
    [$name, $key, $parents] = static::parseName($config_name);
    [$list, $tree] = static::getConfigsList();

    // Get definitions.
    $definition = NestedArray::getValue($tree, $parents);
    if ($definition === NULL) {
      return $local_default;
    }
    $configs = self::populateValues($definition, $local_default);

    return $configs;
  }

  /**
   * Override config value with new one.
   *
   * @param string $config_name
   *   Config name.
   * @param mixed $value
   *   New config value.
   */
  public static function override($config_name, $value) {
    [$name, $key, $parents] = static::parseName($config_name);
    \Drupal::state()->set(static::getNoCollisionName($name), $value);
  }

  /**
   * Override config value with new one.
   *
   * @param string $config_name
   *   Config name.
   * @param mixed $value
   *   New config value.
   */
  public static function overrideSession($config_name, $value) {
    [$name, $key, $parents] = static::parseName($config_name);
    $_SESSION['qtools_common'][static::NO_COLLISION_PREFIX][$name] = $value;
  }

  /**
   * Get session overrides.
   *
   * @return array
   *   Full list of session overrides.
   */
  public static function listSessionOverrides() {
    return $_SESSION['qtools_common'][static::NO_COLLISION_PREFIX] ?? [];
  }

  /**
   * Restore (remove override) of config value.
   *
   * @param string $config_name
   *   Config name.
   */
  public static function restore($config_name) {
    [$name, $key, $parents] = static::parseName($config_name);
    \Drupal::state()->delete(static::getNoCollisionName($name));
    \Drupal::state()->delete($key);
    if (isset($_SESSION['qtools_common'][static::NO_COLLISION_PREFIX][$name])) {
      unset($_SESSION['qtools_common'][static::NO_COLLISION_PREFIX][$name]);
    }
  }

  /**
   * Gets no collision name.
   */
  protected static function getNoCollisionName(string $name): string {
    $names = &drupal_static(static::class . __FUNCTION__, []);
    if (empty($names[$name])) {
      // Config name in state can't be longer than 128 characters.
      // -1 for double-colon.
      $available_length = 128 - strlen(self::NO_COLLISION_PREFIX) - 1;
      if (strlen($name) > $available_length) {
        $hash = sha1($name);
        $name = substr($name . ':' . $hash, -$available_length);
      }
      $names[$name] = static::NO_COLLISION_PREFIX . ':' . $name;
    }
    return $names[$name];
  }

  /**
   * Gets config definition.
   *
   * @param string $config_name
   *   Config name.
   *
   * @return array
   *   Config definition.
   */
  public static function getConfigDefinition($config_name) {
    [$name, $key, $parents] = static::parseName($config_name);
    [$list, $tree] = static::getConfigsList();

    // Get definitions.
    $definition = NestedArray::getValue($tree, $parents);

    return $definition;
  }

  /**
   * Builds config form field base.
   *
   * @param array $definition
   *   Config definition.
   * @param mixed $active_value
   *   Currently active value.
   * @param mixed $fallback_value
   *   Fallback value.
   * @param array $attributes
   *   Form element attributes.
   *
   * @return array
   *   Form field definition.
   */
  protected static function buildConfigFormFieldBase(array $definition, $active_value, $fallback_value, array $attributes) {
    $element = [];

    switch ($definition['type']) {
      case 'double':
        $element['#type'] = 'textfield';
        $element['#element_validate'][] = function (&$element, FormStateInterface $form_state) {
          if (!is_numeric($element['#value'])) {
            $form_state->setError($element, t('%name is not a valid number.', ['%name' => $element['#title']]));
          }
        };
        break;

      case 'integer':
        $element['#type'] = 'number';
        break;

      case 'boolean':
        $element['#type'] = 'select';
        $element['#options'] = [
          "" => t('No (FALSE)'),
          "1" => t('Yes (TRUE)'),
        ];
        break;

      case 'text':
        $element['#type'] = 'textarea';
        break;

      case static::TYPE_ARRAY_MAP:
        $element['#type'] = 'textarea';
        $element['#default_value'] = static::packArrayMap($active_value);
        $attributes['data-qtools-config-fallback-value'] = static::packArrayMap($fallback_value);
        $attributes['rows'] = min(max(count($active_value) + 2, 10), 20);
        break;

      case static::TYPE_ARRAY_LIST:
        $element['#type'] = 'textarea';
        $element['#default_value'] = static::packArrayList($active_value);
        $attributes['data-qtools-config-fallback-value'] = static::packArrayList($fallback_value);
        $attributes['rows'] = min(max(count($active_value) + 2, 10), 20);
        break;

      // String and any unknown ones.
      default:
        $element['#type'] = 'textfield';
        break;

    }

    if (!isset($element['#default_value'])) {
      $element['#default_value'] = $active_value;
    }
    if (!isset($attributes['data-qtools-config-fallback-value'])) {
      $attributes['data-qtools-config-fallback-value'] = (string) $fallback_value;
    }
    $element['#attributes'] = $attributes;

    return $element;
  }

  /**
   * Builds form field.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   * @param string $config_name
   *   Config name.
   * @param array $form_element_info
   *   Override form element props.
   * @param array $definition
   *   Config definition.
   *
   * @return array
   *   Form field definition.
   */
  public static function buildConfigFormField(FormStateInterface $form_state, $config_name, array $form_element_info, array $definition = []) {
    // Load config definition.
    $definition = $definition ?: QToolsConfigHelper::getConfigDefinition($config_name);
    $config_name = $config_name ?? $definition[static::NAME];
    $key = Slug::slugify($config_name);

    // Config name and description.
    $config_title = $config_name;
    $config_description = $config_name;

    // If simple config - build elements.
    if (isset($definition[static::NAME])) {
      // Read states.
      [$states, $active_source] = QToolsConfigHelper::getConfigStates($definition);
      $active_value = $states[$active_source]['value'] ?? NULL;

      // Get fallback value/source.
      if ($active_source != QToolsConfigHelper::SOURCE_STATE) {
        $overridden = FALSE;
        $fallback_source = $active_source;
      }
      else {
        $overridden = TRUE;
        foreach ($states as $type => $info) {
          if ($info['status'] == static::STATE_OVERRIDEN) {
            $fallback_source = $type;
            break;
          }
        }
      }

      // Something went terribly wrong.
      if (empty($fallback_source)) {
        $fallback_source = QToolsConfigHelper::SOURCE_DEFAULT;
      }

      $js_states = [];
      foreach ($states as $type => $info) {
        $js_states[] = ['type' => $type] + $info;
      }

      if ($definition['type'] == 'boolean') {
        $fallback_value = $states[$fallback_source]['value'] ?
          t('Yes (TRUE)') :
          t('No (FALSE)');
      }
      else {
        $fallback_value = $states[$fallback_source]['value'];
      }
      $checkbox_title = t('Fallback value is [@value] taken from [@source]', [
        '@value' => is_array($fallback_value) ? static::packArrayList($fallback_value) : $fallback_value,
        '@source' => $fallback_source,
      ]);
      $payload = [
        'states' => $js_states,
        'title' => $checkbox_title,
      ];
      $attributes = [
        'class' => ['form-element--qtools-common-config'],
        'data-qtools-config-name' => $config_name,
        'data-qtools-config-json' => Json::encode($payload),
        'data-qtools-config-fallback-source' => $fallback_source,
      ];

      $input = static::buildConfigFormFieldBase($definition,
        $active_value,
        $states[$fallback_source]['value'],
        $attributes
      );
      $input += [
        '#qtools_common_config' => $config_name,
        '#title' => $config_title,
        '#description' => $config_description,
        '#attached' => [
          'library' => 'qtools_common/form-element--qtools-common-config',
        ],
      ];

      $override_enabled = [
        '#type' => 'hidden',
        '#default_value' => (int) $overridden,
      ];

      $element = [
        $key => $form_element_info + $input,
        $key . static::KEY_SUFFIX_OVERRIDE_ENABLED => $override_enabled,
      ];
    }
    elseif (!empty($definition)) {
      // Create a fieldset.
      $element = [
        '#type' => 'fieldset',
        '#title' => $config_name,
      ];
      foreach ($definition as $key => $sub_def) {
        [$sub_key, $sub_element] = static::buildConfigFormField($form_state, NULL, [], $sub_def);
        $element[$sub_key] = $sub_element;
      }
    }
    else {
      // Definition not found.
      $input = [
        '#type' => 'item',
        '#title' => $config_title,
        '#description' => $config_description,
        '#markup' => t('<em>!!! Config definition not found, please report to site administrator !!!</em>'),
      ];
      $element = [
        $key => $input,
      ];
    }

    // Save mapping between form keys and config names.
    $config_keys_map = $form_state->getTemporaryValue('qtools_common_config_keys_map') ?? [];
    $config_keys_map[$config_name] = $key;
    $form_state->setTemporaryValue('qtools_common_config_keys_map', $config_keys_map);

    return [$key . static::KEY_SUFFIX_GROUP, $element];
  }

  /**
   * Submit configuration work based on form state values.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   */
  public static function submitConfigurationForm(FormStateInterface $form_state) {
    // Read map of config fields on a form.
    $config_keys_map = $form_state->getTemporaryValue('qtools_common_config_keys_map') ?? [];
    foreach ($config_keys_map as $config_name => $form_key) {
      // Read override status from raw values.
      $overridden_key = $form_key . static::KEY_SUFFIX_OVERRIDE_ENABLED;
      $override_enabled = $form_state->getValue($overridden_key) ?? "0";
      if (!empty($override_enabled)) {
        $definition = QToolsConfigHelper::getConfigDefinition($config_name);

        // Read value from form state.
        $form_value = $form_state->getValue($form_key);
        if ($definition['type'] == static::TYPE_ARRAY_MAP) {
          $value = static::unpackArrayMap($form_value);
        }
        elseif ($definition['type'] == static::TYPE_ARRAY_LIST) {
          $value = static::unpackArrayList($form_value);
        }
        else {
          $value = $form_value;
        }

        static::override($config_name, $value);
      }
      else {
        static::restore($config_name);
      }
    }
  }

}
