<?php

namespace Drupal\qtools_common;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\Site\Settings;

/**
 * Class QToolsCryptHelper.
 *
 * @package Drupal\qtools
 */
abstract class QToolsCryptHelper {

  const SIGN_FIXED = 0;
  const SIGN_SITE = 1;
  const SIGN_USER = 2;
  const SIGN_SESSION = 3;

  const CHECK_VALID = 0;
  const CHECK_MALFORMED = -1;
  const CHECK_EXPIRED = -2;
  const CHECK_INVALID = -3;

  const ENCRYPTION_CIPHER = 'aes-256-ctr';
  const ENCRYPTION_HASH_LENGTH = 32;

  /**
   * Encrypts string message.
   *
   * @see https://www.php.net/manual/ru/function.openssl-encrypt.php
   *
   * @return string|null
   *   Encrypted message.
   */
  public static function encrypt($plaintext, $key, $cipher = self::ENCRYPTION_CIPHER) {
    // Check if encryption method available.
    if (in_array($cipher, openssl_get_cipher_methods())) {
      $ivlen = openssl_cipher_iv_length($cipher);
      $iv = openssl_random_pseudo_bytes($ivlen);
      $ciphertext_raw = openssl_encrypt($plaintext, $cipher, $key, OPENSSL_RAW_DATA, $iv);
      $hmac = hash_hmac('sha256', $ciphertext_raw, $key, TRUE);
      $ciphertext = base64_encode($iv . $hmac . $ciphertext_raw);
    }
    return $ciphertext ?? NULL;
  }

  /**
   * Decrypt string message (only works with data encrypted with QToolsCryptHelper::encrypt).
   *
   * @see https://www.php.net/manual/ru/function.openssl-encrypt.php
   *
   * @return string|null
   *   Decrypted message.
   */
  public static function decrypt($ciphertext, $key, $cipher = self::ENCRYPTION_CIPHER) {
    // Check if encryption method available.
    if (in_array($cipher, openssl_get_cipher_methods())) {
      $decoded = base64_decode($ciphertext);
      $ivlen = openssl_cipher_iv_length($cipher);
      $iv = substr($decoded, 0, $ivlen);
      $hmac = substr($decoded, $ivlen, static::ENCRYPTION_HASH_LENGTH);
      $ciphertext_raw = substr($decoded, $ivlen + static::ENCRYPTION_HASH_LENGTH);
      $original_plaintext = openssl_decrypt($ciphertext_raw, $cipher, $key, OPENSSL_RAW_DATA, $iv);
      $calcmac = hash_hmac('sha256', $ciphertext_raw, $key, TRUE);
      if (hash_equals($hmac, $calcmac)) {
        return $original_plaintext;
      }
    }
    return NULL;
  }

  /**
   * Checks signed string for validity and extract payload.
   *
   * @param string $signed_string
   *   Signed string.
   * @param mixed $data
   *   Payload data part.
   *
   * @return int
   *   Check code.
   */
  public static function check($signed_string, &$data = NULL) {

    // Split into payload and hash.
    $parts = explode('.', (string) $signed_string);
    if (count($parts) != 2) {
      return static::CHECK_MALFORMED;
    }
    $payload_decoded = base64_decode($parts[0]);
    $payload = (array) json_decode($payload_decoded);
    $data = isset($payload['d']) ? $payload['d'] : NULL;

    // Validate payload.
    if (count($payload) != 3 || !isset($payload['d']) || !isset($payload['e']) || !isset($payload['t'])) {
      return static::CHECK_MALFORMED;
    }

    // Validate hash.
    $new_hash_string = $parts[0] . self::getSeed($payload['t']);
    $new_hash = Crypt::hashBase64($new_hash_string);
    if ($new_hash !== $parts[1]) {
      return static::CHECK_INVALID;
    }

    // Validate timestamp (if not permanent).
    if ($payload['e'] != 0 && $payload['e'] < time()) {
      return static::CHECK_EXPIRED;
    }

    // If we get here we have a valid.
    return static::CHECK_VALID;
  }

  /**
   * Sign some data.
   *
   * If element is missing key property/key it will be removed
   * from resulted array.
   *
   * @param mixed $data
   *   Data to sign.
   * @param int $type
   *   Type of signature.
   * @param int $ttl
   *   Optional sign ttl.
   *
   * @return string
   *   Base64 string signed.
   */
  public static function sign($data, $type, $ttl = 0) {

    $expire = !empty($ttl) ? time() + $ttl : 0;

    $payload = [
      'd' => $data,
      'e' => $expire,
      't' => $type,
    ];

    $payload_string = base64_encode(json_encode($payload));
    $hash_string = $payload_string . self::getSeed($type);
    $hash = Crypt::hashBase64($hash_string);
    return $payload_string . '.' . $hash;
  }

  /**
   * Returns appropriate seed based on signature type.
   *
   * @param int $type
   *   Type of signature.
   *
   * @return string
   *   Signature seed.
   */
  protected static function getSeed($type) {
    $seed = Settings::getHashSalt();
    switch ($type) {
      case static::SIGN_SITE:
        // We already have site signature as base.
        break;

      case static::SIGN_USER:
        if (\Drupal::currentUser()->id() !== 0) {
          $seed .= \Drupal::currentUser()->id();
        }
        else {
          // If we did not start session this token will never be valid.
          if (empty($_SESSION['qtools_common_session_started'])) {
            $_SESSION['qtools_common_session_started'] = 1;
          }
          $seed .= session_id();
        }

        break;

      case static::SIGN_SESSION:
        // If we did not start session this token will never be valid.
        if (empty($_SESSION['qtools_common_session_started'])) {
          $_SESSION['qtools_common_session_started'] = 1;
        }

        $seed .= session_id();
        break;

      default:
        // If no type specified make sure this signature cannot be verified.
        $seed .= uniqid('', TRUE);
        break;
    }
    return $seed;
  }

}
