<?php

namespace Drupal\qtools_common;

use Drupal\qtools_common\Utility\Slug;

/**
 * Class QToolsDatasetHelper.
 *
 * Provides means to query PHP arrays mimicking database flow.
 *
 * @package Drupal\qtools
 */
abstract class QToolsDatasetHelper {

  const PROP_TYPE_OPTIONS = 'o';

  const INDEX_DEFAULT = '_';

  const FILTER_OP_IN = 'IN';

  // Facet modes, note that facets are performance taxing,
  // especially inclusive ones.
  const FACET_MODE_NONE = 0;
  const FACET_MODE_EXCLUSIVE = 1;
  const FACET_MODE_INCLUSIVE = 2;

  /**
   * Filters single index against aligned filters.
   *
   * @param array $index
   *   Index to filter.
   * @param array $aligned_filters
   *   Filters aligned to dataset.
   * @param array $props_for_facets
   *   If props provided, facets will be collected.
   *
   * @return array
   *   Filtered index with facets [$result, $facets].
   */
  protected static function filterIndex(array $index, array $aligned_filters, array $props_for_facets = NULL) {
    $result = [];
    $facets = [];
    foreach ($index as $record_id => $record_props) {
      $facet_prop_id = NULL;
      if (static::propsMatchFilters($record_props, $aligned_filters, $facet_prop_id)) {
        $result[] = $record_id;

        // Collect normal facets.
        if (!empty($props_for_facets)) {
          foreach ($record_props as $prop_id => $value) {
            if ($value !== NULL && $props_for_facets[$prop_id]['facet_mode'] !== static::FACET_MODE_NONE) {
              $facets[$prop_id][$value] = !empty($facets[$prop_id][$value]) ?
                $facets[$prop_id][$value] + 1 :
                1;
            }
          }
        }
      }
      elseif (!empty($props_for_facets) && !empty($facet_prop_id)) {
        // Collect facets from mismatches, only applicable for inclusive mode.
        $facet_prop_value = $record_props[$facet_prop_id];
        if (!empty($facets[$facet_prop_id][$facet_prop_value])) {
          $facets[$facet_prop_id][$facet_prop_value]++;
        }
        else {
          $facets[$facet_prop_id][$facet_prop_value] = 1;
        }
      }
    }
    return [$result, $facets];
  }

  /**
   * Checks if the given props array matches filters.
   *
   * @param array $record_props
   *   Array or normalized props.
   * @param array $aligned_filters
   *   Array of aligned filters.
   * @param mixed $facet_prop_id
   *   Will contain prop id that contributes to facets
   *   in case of mismatched record.
   *
   * @return bool
   *   TRUE if match, FALSE if does not.
   */
  protected static function propsMatchFilters(array $record_props, array $aligned_filters, &$facet_prop_id = NULL) {
    $match = TRUE;
    $facet_prop_id = NULL;
    foreach ($aligned_filters as $prop_id => $filter) {
      switch ($filter['op']) {
        case static::FILTER_OP_IN:
          $record_prop_value = $record_props[$prop_id] ?? NULL;

          // Get record value for given prop, if empty
          // then it does not influence facets and is a MISSMATCH.
          // So no point in continue filtering.
          if ($record_prop_value === NULL) {
            $facet_prop_id = FALSE;
            return FALSE;
          }
          elseif (!in_array($record_prop_value, $filter['value'], TRUE)) {
            $match = FALSE;
            // If filter isn't matched, it still CAN contribute
            // to its respective facet, if it would match other filters
            // and facet_mode is inclusive.
            if ($facet_prop_id === NULL && $filter['facet_mode'] == static::FACET_MODE_INCLUSIVE) {
              $facet_prop_id = $prop_id;
            }
            else {
              $facet_prop_id = FALSE;

              // At this point record cant neither match or contribute
              // to facets so no point, so no point in continue filtering.
              return FALSE;
            }
          }
          break;
      }
    }
    return $match;
  }

  /**
   * Query dataset against filters.
   *
   * @param array $dataset
   *   Dataset to query from.
   * @param array $filters
   *   Normalised filters.
   * @param bool $ignore_unknown_filters
   *   If TRUE, any filter that is not part of dataset will be ignored,
   *   otherwise search will always return 0 results, so no point in search.
   * @param bool $collect_facets
   *   If TRUE, facets will be collected during query.
   * @param int $size
   *   Page size if paged output.
   * @param int $page
   *   Page number if paged output (starts with 0).
   *
   * @return array
   *   Query result array [$result, $pager, $facets].
   */
  public static function queryDataset(array $dataset, array $filters = [], $ignore_unknown_filters = TRUE, $collect_facets = FALSE, $size = NULL, $page = 0) {
    $index = $dataset['indexes'][static::INDEX_DEFAULT] ?? [];

    // Align filters.
    $aligned_filters = static::alignFiltersWithDataset($filters, $dataset, $ignore_unknown_filters);

    // Broken filters = no results.
    if ($aligned_filters === NULL) {
      $result = [];
      $facets = [];
    }
    elseif ($aligned_filters === []) {
      // Empty filters = all results.
      $result = array_keys($index);
      $facets = $dataset['facets'];
    }
    else {
      // Actually filter dataset.
      $props_for_facets = $collect_facets ? $dataset['props_by_id'] : NULL;
      [$result, $facets] = static::filterIndex($index, $aligned_filters, $props_for_facets);
    }

    // Pager support.
    $pager['total_items'] = count($result);

    if (!empty($size)) {
      $pages = array_chunk($result, $size);
      $pager['total_pages'] = count($pages);
      $result = $pages[$page] ?? [];

    }
    else {
      $pager['total_pages'] = $pager['total_items'] > 0 ? 1 : 0;
    }
    $pager['size'] = $size;
    $pager['page'] = $page;

    return [$result, $pager, $facets];
  }

  /**
   * Align normalised filters (by key) to dataset (by id).
   *
   * @param array $filters
   *   Array of normalised filters.
   * @param array $dataset
   *   Dataset to align to.
   * @param bool $ignore_unknown_filters
   *   If true, any filter that is not part of dataset will be ignored,
   *   otherwise search will always return 0 results, so no point in search.
   *
   * @return array|null
   *   Aligned filters or null if filter not present in dataset and not ignored.
   */
  protected static function alignFiltersWithDataset(array $filters, array $dataset, $ignore_unknown_filters) {
    $aligned = [];
    foreach ($filters as $prop_key => $filter) {
      // Check if filter exists in dataset.
      if (empty($dataset['props_by_key'][$prop_key])) {
        // If we don't ignore unknown filters, result will
        // always be empty so no point in search.
        if (!$ignore_unknown_filters) {
          return NULL;
        }
        continue;
      }
      $prop = $dataset['props_by_key'][$prop_key];

      if ($prop['type'] == static::PROP_TYPE_OPTIONS) {
        // Operation for options filters in always 'IN'.
        $aligned[$prop['id']]['op'] = static::FILTER_OP_IN;
        $aligned[$prop['id']]['facet_mode'] = $prop['facet_mode'];
        foreach ($filter as $value_key) {
          if (!empty($prop['values_by_key'][$value_key])) {
            $aligned[$prop['id']]['value'][] = $prop['values_by_key'][$value_key]['id'];
          }
          elseif (!$ignore_unknown_filters) {
            // If we don't ignore unknown filters, result will
            // always be empty so no point in search.
            return NULL;
          }
        }
      }
    }

    return $aligned;
  }

  /**
   * Default key normalisation function.
   *
   * @param string|int $value
   *   Original key.
   *
   * @return string
   *   Normalised key.
   */
  public static function defaultKeyTransform($value) {
    return Slug::slugify($value);
  }

  /**
   * Creates dataset from keyed array.
   *
   * @param array $keyed_table
   *   Table in $key => ['prop1', 'prop2'...] format.
   * @param array $props_header
   *   Properties headers (names).
   * @param mixed $null_value
   *   Value that will be treated as null.
   * @param callable $name_transform
   *   Name transformation callable.
   * @param callable $key_transform
   *   Key transormation function, if empty
   *   static:defaultKeyTransform() will be used.
   *
   * @return array
   *   Dataset array.
   */
  public static function createFromKeyedTable(array $keyed_table, array $props_header = [], $null_value = NULL, callable $name_transform = NULL, callable $key_transform = NULL) {

    $indexes = [];
    $facets = [];

    // Create props map.
    $props_by_id = [];
    $props_by_key = [];
    foreach ($props_header as $prop_pos => $prop_info) {
      if (is_array($prop_info)) {
        $prop_name = $prop_info['name'];
        $prop_facet_mode = $prop_info['facet_mode'];
      }
      else {
        $prop_name = $prop_info;
        $prop_facet_mode = static::FACET_MODE_INCLUSIVE;
      }

      // Start from 1 to avoid inconsistent comparison.
      $prop_id = $prop_pos + 1;
      $prop_key = is_callable($key_transform) ?
        $key_transform($prop_name) :
        static::defaultKeyTransform($prop_name);
      $prop_info = [
        'id' => $prop_id,
        'key' => $prop_key,
        'name' => $prop_name,
        'type' => static::PROP_TYPE_OPTIONS,
        'facet_mode' => $prop_facet_mode,
        'map' => [],
      ];
      $props_by_id[$prop_id] = $prop_info;
      $props_by_key[$prop_key] = &$props_by_id[$prop_id];
    }

    // Build default index.
    $props_compact = [];
    foreach ($keyed_table as $item_key => $props_list) {
      $props = [];
      foreach ($props_list as $prop_pos => $value) {
        $prop_id = $prop_pos + 1;

        // Null properties are not present in any maps.
        if ($value === $null_value) {
          $props[$prop_id] = NULL;
          continue;
        }

        // Normalise value.
        if ($props_by_id[$prop_id]['type'] == static::PROP_TYPE_OPTIONS) {
          $value_key = is_callable($key_transform) ?
            $key_transform($value) :
            static::defaultKeyTransform($value);
          $value_name = is_callable($name_transform) ?
            $name_transform($value) :
            $value;

          // Check if this value already present in the map.
          if (!isset($props_compact[$prop_id][$value_key])) {
            // Start from 1 to avoid inconsistent comparison.
            $value_id = !empty($props_compact[$prop_id]) ?
              count($props_compact[$prop_id]) + 1 :
              1;
            $props_compact[$prop_id][$value_key] = [
              'id' => $value_id,
              'name' => $value_name,
            ];
          }
          else {
            $value_id = $props_compact[$prop_id][$value_key]['id'];
          }

          $props_by_id[$prop_id]['values_by_id'][$value_id] = [
            'id' => $value_id,
            'key' => $value_key,
            'name' => $value_name,
          ];
          $props_by_id[$prop_id]['values_by_key'][$value_key] = &$props_by_id[$prop_id]['values_by_id'][$value_id];
        }
        else {
          $value_id = $value;
        }

        $props[$prop_id] = $value_id;
        $facets[$prop_id][$value_id] = !empty($facets[$prop_id][$value_id]) ?
          $facets[$prop_id][$value_id] + 1 :
          1;
      }
      $indexes[static::INDEX_DEFAULT][$item_key] = $props;
    }

    return [
      'indexes' => $indexes,
      'facets' => $facets,
      'props_by_id' => $props_by_id,
      'props_by_key' => $props_by_key,
    ];

  }

}
