<?php

namespace Drupal\qtools_common;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\views\ViewExecutable;
use Drupal\views\Views;

/**
 * Class QToolsViewsHelper.
 *
 * @package Drupal\qtools
 */
abstract class QToolsViewsHelper {

  /**
   * Get view's real build array.
   */
  public static function getViewContent(string $name, array $args = [], array $exposed = [], mixed $items_per_page = FALSE, string $display_id = 'default', ?RefinableCacheableDependencyInterface $cacheable_metadata = NULL): array {
    // Get prepared view object.
    $view = static::getPreparedView($name, $args, $exposed, $items_per_page, $display_id);
    if (empty($view)) {
      // If view was not found or some other error appears we don't cache this result.
      $cacheable_metadata?->mergeCacheMaxAge(QToolsCacheHelper::MAX_AGE_NO_CACHE);
      return [];
    }

    $build = $view->render();

    if (!empty($cacheable_metadata)) {
      $cacheable_metadata->addCacheableDependency(CacheableMetadata::createFromRenderArray($build));
    }

    return $build;
  }

  /**
   * Get view's result from input parameters.
   *
   * @return \Drupal\views\ResultRow[]
   *   View execution result.
   */
  public static function getViewResult(string $name, array $args = [], array $exposed = [], mixed $items_per_page = FALSE, string $display_id = 'default', ?RefinableCacheableDependencyInterface $cacheable_metadata = NULL): array {
    // Get prepared view object.
    $view = static::getPreparedView($name, $args, $exposed, $items_per_page, $display_id);
    if (empty($view)) {
      // If view was not found or some other error appears we don't cache this result.
      $cacheable_metadata?->mergeCacheMaxAge(QToolsCacheHelper::MAX_AGE_NO_CACHE);
      return [];
    }
    $view->execute();
    $result = $view->result;

    // Add view's cache tags metadata.
    $cacheable_metadata?->addCacheableDependency($view->getDisplay()->getCacheMetadata());
    $cacheable_metadata?->addCacheTags($view->getCacheTags());

    // Manually add metadata from cache plugin.
    /** @var \Drupal\views\Plugin\views\cache\CachePluginBase $cache_plugin */
    $cache_plugin = $view->getDisplay()->getPlugin('cache');
    if (!empty($cache_plugin)) {
      $cacheable_metadata?->mergeCacheMaxAge($cache_plugin->getCacheMaxAge());
      $cacheable_metadata?->addCacheTags($cache_plugin->getCacheTags());
    }

    // Delete view object to optimise memory footprint.
    $view->destroy();

    return $result;
  }

  /**
   * Load view and prepare it for execution.
   */
  protected static function getPreparedView(string $name, array $args = [], array $exposed = [], mixed $items_per_page = FALSE, string $display_id = 'default'): ?ViewExecutable {
    // Load view object.
    $view = Views::getView($name);

    if (empty($view)) {
      return NULL;
    }

    // Set arguments and exposed input.
    if (!empty($args)) {
      $view->setArguments($args);
    }

    if (!empty($exposed)) {
      $view->setExposedInput($exposed);
    }

    // Set items per page.
    if ($items_per_page !== FALSE) {

      // If items per page is array
      // consider first item as offset.
      if (is_array($items_per_page)) {
        $view->setOffset($items_per_page[0]);
        $view->setItemsPerPage($items_per_page[1]);
      }
      else {
        $view->setItemsPerPage($items_per_page);
      }
    }

    if (is_string($display_id)) {
      $view->setDisplay($display_id);
    }
    else {
      $view->initDisplay();
    }

    $view->preExecute();

    return $view;
  }

}
