<?php

namespace Drupal\qtools_common\Utility;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use ArrayObject as ArrayObject;

/**
 * Class CacheableArray.
 */
class CacheableArray extends ArrayObject implements RefinableCacheableDependencyInterface {

  /**
   * Drupal\Core\Cache\CacheableMetadata.
   *
   * @var \Drupal\Core\Cache\CacheableMetadata
   */
  protected $cacheableMetadata;

  /**
   * CacheableArray constructor.
   */
  public function __construct(array $objects = [], CacheableDependencyInterface $cacheableInfo = NULL) {
    // Build Array Object.
    parent::__construct($objects);

    $this->cacheableMetadata = new CacheableMetadata();
    if ($cacheableInfo) {
      $this->cacheableMetadata->addCacheableDependency($cacheableInfo);
    }
  }

  /**
   * Get current cacheable metadata.
   */
  public function getCacheableMetadata(): CacheableMetadata {
    // Base metadata.
    $cacheableMetadata = new CacheableMetadata();
    $cacheableMetadata->addCacheableDependency($this->cacheableMetadata);

    // Merge base metadata with collection metadata.
    foreach ($this as $item) {
      if ($item instanceof CacheableDependencyInterface) {
        $cacheableMetadata->addCacheableDependency($item);
      }
    }
    return $cacheableMetadata;
  }

  /**
   * Slice current array.
   */
  public function slice(int $start, ?int $length = NULL): static {
    $array = $this->getArrayCopy();
    $array = array_slice($array, $start, $length);
    return new static($array, $this->cacheableMetadata);
  }

  /**
   * CacheableDependencyInterface, returns object cache tags.
   */
  public function getCacheTags(): array {
    return $this->getCacheableMetadata()->getCacheTags();
  }

  /**
   * CacheableDependencyInterface, returns object cache contexts.
   */
  public function getCacheContexts(): array {
    return $this->getCacheableMetadata()->getCacheContexts();
  }

  /**
   * CacheableDependencyInterface, returns object cache max age.
   */
  public function getCacheMaxAge(): int {
    return $this->getCacheableMetadata()->getCacheMaxAge();
  }

  /**
   * RefinableCacheableDependencyInterface: addCacheableDependency.
   */
  public function addCacheableDependency($other_object): static {
    $this->cacheableMetadata->addCacheableDependency($other_object);
    return $this;
  }

  /**
   * RefinableCacheableDependencyInterface: addCacheContexts.
   */
  public function addCacheContexts(array $cache_contexts): static {
    if ($cache_contexts) {
      $this->cacheableMetadata->addCacheContexts($cache_contexts);
    }
    return $this;
  }

  /**
   * RefinableCacheableDependencyInterface: addCacheTags.
   */
  public function addCacheTags(array $cache_tags): static {
    if ($cache_tags) {
      $this->cacheableMetadata->addCacheTags($cache_tags);
    }
    return $this;
  }

  /**
   * RefinableCacheableDependencyInterface: mergeCacheMaxAge.
   */
  public function mergeCacheMaxAge($max_age): static {
    $this->cacheableMetadata->mergeCacheMaxAge($max_age);
    return $this;
  }

}
