<?php

namespace Drupal\qtools_common\Utility;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormState;
use Drupal\Core\Form\FormStateInterface;
use Drupal\qtools_common\QToolsConfigHelper;
use Drupal\qtools_common\QToolsConfigurationChangesTrackingService;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for connection settings form.
 */
abstract class QToolsConfigFormBase extends FormBase {

  /**
   * Drupal\qtools_common\QToolsConfigurationChangesTrackingService.
   *
   * @var \Drupal\qtools_common\QToolsConfigurationChangesTrackingService
   */
  protected $changesTrackingService;

  // Key to store config settings.
  const CONFIG_SETTINGS_KEY = '#config_settings_data';

  /**
   * {@inheritdoc}
   */
  public function __construct(
    QToolsConfigurationChangesTrackingService $changesTrackingService
  ) {
    $this->changesTrackingService = $changesTrackingService;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('qtools_common.configuration_changes_tracking')
    );
  }

  /**
   * Inject configuration field in the given place.
   *
   * @return array
   *   Injected element.
   */
  public function addConfigField(&$root, FormStateInterface $form_state, $config_name, $form_element_info) {
    // Sometimes element info is just a title.
    if (!is_array($form_element_info)) {
      $form_element_info = ['#title' => $form_element_info];
    }
    [$element_key, $config_element] = QToolsConfigHelper::buildConfigFormField($form_state, $config_name, $form_element_info);
    $root[$element_key] = $config_element;
    return $config_element;
  }

  /**
   * Get list of sections with config params, see example function.
   */
  abstract public function getSections(): array;

  /**
   * Specify if this form changes should be tracked (disabled by default).
   *
   * @return bool
   *   TRUE if this form changes should be tracked.
   */
  abstract public function trackingEnabled();

  /**
   * Returns validated and connected config sections.
   */
  public function getSectionsExample() {
    $sections['section1'] = [
      'title' => $this->t("Section name 1"),
      'description' => $this->t('Section description 1'),
      'form_elements' => [
        "sg_cart.order.max_distance" => $this->t('Simple element 1'),
        "sg_cart.order.max_distance.stockSource" => $this->t('Simple element 2'),
        "sg_common.map.locations_radius" => $this->t("Simple element 3"),
      ],
    ];

    $sections['section2'] = [
      'title' => $this->t("Section name 2"),
      'description' => $this->t('Section description 2'),
      'form_elements' => [
        "sg_site.maintenance_mode_url_redirect" => $this->t('Simple element 1'),
      ],
    ];

    return $sections;
  }

  /**
   * Builds form, adding default elements.
   *
   * @return array
   *   Form.
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    // Read sections.
    $sections = $this->getSections();

    $form['instructions'] = [
      '#type' => '#markup',
      '#markup' => 'Guidelines:</br><ul><li>Click checkbox in front of a setting to override it.</li><li>Mouse over checkbox to see fallback value and where its taken from.</li></ul>',
    ];

    $form['information'] = [
      '#type' => 'vertical_tabs',
      '#default_tab' => 'edit-' . array_keys($sections)[0],
    ];

    // Process all config sections.
    $flat_list = [];
    foreach ($sections as $section_id => $section_info) {
      $form[$section_id] = [
        '#type' => 'details',
        '#title' => $section_info['title'],
        '#description' => $section_info['description'],
        '#group' => 'information',
      ];
      $flat_part = $this->buildFormElements($form, $form_state, $section_info, $section_id);
      $flat_list = array_merge($flat_list, $flat_part);
    }
    $form['#flat_list'] = $flat_list;

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => 'save',
    ];
    return $form;
  }

  /**
   * Add elements.
   */
  protected function buildFormElements(&$form, FormStateInterface $form_state, $section_info, $section_id) {
    $flat_list = [];
    foreach ($section_info['form_elements'] as $config_name => $element_info) {
      if (is_array($element_info) && empty($element_info['#type'])) {
        $form[$section_id][$config_name] = [
          '#type' => 'fieldset',
          '#title' => $element_info['#title'],
        ];
        foreach ($element_info as $sub_config_name => $sub_element_info) {
          if (!str_contains($sub_config_name, '#')) {
            $flat_list[$sub_config_name] = $this->addConfigField($form[$section_id][$config_name], $form_state, $sub_config_name, $sub_element_info);
          }
        }
      }
      else {
        $flat_list[$config_name] = $this->addConfigField($form[$section_id], $form_state, $config_name, $element_info);
      }
    }
    return $flat_list;
  }

  /**
   * Submits a form, saving/resetings all config values.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    QToolsConfigHelper::submitConfigurationForm($form_state);

    // Track changes if enabled.
    if ($this->trackingEnabled()) {
      // Rebuild form to get new values.
      $updated_form = [];
      $updated_form_state = new FormState();
      $updated_form = $this->buildForm($updated_form, $updated_form_state);

      $this->trackChanges($form['#flat_list'], $updated_form['#flat_list']);
    }

    $this->messenger()->addStatus($this->t('The changes have been saved.'));
  }

  /**
   * Tracks changes to the form.
   */
  protected function trackChanges($flat_list, $updated_flat_list) {
    $diff = [];
    foreach ($flat_list as $config_key => $config_fields) {
      $key = Slug::slugify($config_key);
      $overridden_key = $key . QToolsConfigHelper::KEY_SUFFIX_OVERRIDE_ENABLED;

      $value = $config_fields[$key]['#default_value'];
      $overridden = $config_fields[$overridden_key]['#default_value'];
      $updated_value = $updated_flat_list[$config_key][$key]['#default_value'];
      $updated_overridden = $updated_flat_list[$config_key][$overridden_key]['#default_value'];

      if ($value !== $updated_value || $overridden !== $updated_overridden) {

        // For values with options, use human-readable variants.
        $text_value = !empty($config_fields[$key]['#options']) ?
          $config_fields[$key]['#options'][(string) $value] :
          $value;

        $text_overridden_value = !empty($updated_flat_list[$config_key][$key]['#options']) ?
          $updated_flat_list[$config_key][$key]['#options'][(string) $updated_value] :
          $updated_value;

        $diff[$config_key] = [
          $text_value,
          $overridden,
          $text_overridden_value,
          $updated_overridden,
        ];
      }
    }

    foreach ($diff as $config_key => $config_change) {
      [$value, $overridden, $updated_value, $updated_overridden] = $config_change;
      $this->changesTrackingService->trackConfigChange($config_key, $value, $overridden, $updated_value, $updated_overridden);
    }
  }

}
