<?php

namespace Drupal\qtools_common\Utility;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\filter\Render\FilteredMarkup;
use Symfony\Component\VarDumper\Cloner\Data;
use Symfony\Component\VarDumper\Cloner\VarCloner;
use Symfony\Component\VarDumper\Dumper\CliDumper;
use Symfony\Component\VarDumper\Dumper\HtmlDumper;

/**
 * Provides helper to debug.
 *
 * @ingroup utility
 */
class QToolsDebug {

  /**
   * Checks for recursion and die after some amount of invocations.
   */
  public static function dieAt($die_count = 1) {
    $count = &drupal_static('QToolsDebug:stopAt', 0);
    $count++;

    if ($count >= $die_count) {
      $func = 'die';
      $func('died on ' . $die_count);
    }
  }

  /**
   * Prints debug message taking into account place it was called.
   */
  public static function log($data = 'called', $name = '', $call_step = 1) {
    $suffix = static::getDbtCallerString($call_step + 1);
    static::message($data, $name, $suffix);
  }

  /**
   * Get debug backtrace caller string to this invocation.
   */
  public static function getDbtCallerString($call_step = 1) {
    // Get called class name.
    $bt = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
    $call = $bt[$call_step];
    $call_1 = $bt[$call_step + 1];
    return ($call['class'] ?? '') . ($call['type'] ?? '') . $call_1['function'] . '()  @ ' . $call['file'] . ':' . $call['line'];
  }

  /**
   * Prints debug message taking into account place it was called.
   */
  public static function message(mixed $input, ?string $name = NULL, ?string $suffix = NULL) {
    $output = static::export(static::cloneFormVarDumper($input), $name, $suffix);
    \Drupal::messenger()->addMessage($output, MessengerInterface::TYPE_STATUS, TRUE);
  }

  /**
   * Starts/measures timer and return its key.
   */
  public static function timer($key = NULL, $log = TRUE, $call_step = 2) {
    $timers = &drupal_static('QToolsDebug:timers', []);

    // Create new timer if no key passed.
    if ($key === NULL) {
      $key = uniqid(TRUE);
      $timers[$key] = microtime(TRUE);
      return $key;
    }
    else {
      // Return elapsed time on given timer.
      $elapsed = microtime(TRUE) - $timers[$key];
      if ($log) {
        static::log($elapsed, $log, $call_step);
      }
      return $elapsed;
    }
  }

  /**
   * Clone variable to be used in "dpm" like format (Symfony VarDumper).
   */
  public static function cloneFormVarDumper(mixed $input): ?Data {
    if ($input instanceof Data) {
      return $input;
    }
    if (class_exists('\Symfony\Component\VarDumper\Cloner\VarCloner')) {
      $casters = VarCloner::$defaultCasters;

      // Add caster for translatable markup.
      $casters['Drupal\Core\StringTranslation\TranslationManager'] = function ($item) {
        return [
          '::class' => get_class($item),
        ];
      };

      $cloner = new VarCloner($casters);
      return $cloner->cloneVar($input);
    }
    else {
      return NULL;
    }
  }

  /**
   * Exports variable using "dpm" like format (Symfony VarDumper).
   */
  public static function export(?Data $data = NULL, ?string $name = NULL, ?string $suffix = NULL): MarkupInterface|string {
    if (!empty($data)) {
      $dumper = 'cli' === PHP_SAPI ? new CliDumper() : new HtmlDumper();
      $output = fopen('php://memory', 'r+b');
      $dumper->dump($data, $output);
      $output = stream_get_contents($output, -1, 0);
    }
    else {
      $output = 'NULL';
    }

    if ($name) {
      $output = $name . ' => ' . $output;
    }
    if ($suffix) {
      $output .= $suffix;
    }

    return FilteredMarkup::create($output);
  }

}
