<?php

namespace Drupal\qtools_common\Utility;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Psr\Log\LoggerInterface;

/**
 * Generic class with cacheable metadata + errors management.
 */
class ResultMetadata extends CacheableMetadata {

  /**
   * Logger channel name, if omitted then logs are not automatically written.
   *
   * @var string
   */
  protected string $loggerChannel;

  /**
   * Metadata storage.
   *
   * @var array
   */
  protected array $state = [];

  /**
   * Logger.
   *
   * @var \Psr\Log\LoggerInterface|null
   */
  protected ?LoggerInterface $logger = NULL;

  /**
   * Messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface|null
   */
  private ?MessengerInterface $messenger;

  /**
   * Constructs new result metadata object.
   */
  public function __construct(?LoggerInterface $logger = NULL, ?MessengerInterface $messenger = NULL) {
    $this->logger = $logger;
    $this->messenger = $messenger;
  }

  /**
   * Provides translation capabilities.
   *
   * Can't use StringTranslation trait, as it adds database connection
   * dependent service, which makes this and descendent objects
   * non-serializable.
   */
  protected function t($string, array $args = [], array $options = []): TranslatableMarkup {
    return t($string, $args, $options);
  }

  /**
   * Adds error to metadata pool, logs it into watchdog if logger channel specified.
   */
  public function addError($message, ?string $key = NULL, $debug_data = NULL): void {
    if (!empty($this->logger)) {
      $this->logger->error($message);
    }
    if (!empty($this->messenger)) {
      $this->messenger->addError($message);
    }

    // Only store serializable data.
    if ($message instanceof TranslatableMarkup) {
      $message = $message->render();
    }

    $this->state['errors'][] = [$message, $key, $debug_data];
  }

  /**
   * Merge errors from another resultMetadata.
   */
  public function extendWith(self $resultMetadata): void {
    // Extend cacheable metadata.
    $this->addCacheableDependency($resultMetadata);

    // Extend errors.
    $errors = $resultMetadata->getState('errors');

    if (empty($errors)) {
      return;
    }

    foreach ($errors as $error) {
      $this->addError($error[0], $error[1], $error[2]);
    }
  }

  /**
   * Returns TRUE if metadata contains error.
   */
  public function hasErrors(): bool {
    return !empty($this->state['errors']);
  }

  /**
   * Returns TRUE if no errors, usecase is to combine with NULL checks.
   *
   * Example use case if ($resultMetadata?->isValid()) {....}
   * This will fail if there are errors OR if $resultMetadata was NULL.
   */
  public function isValid(): bool {
    return !$this->hasErrors();
  }

  /**
   * Returns all errors without debug data.
   */
  public function getErrors(): array {
    $errors = [];
    if (!empty($this->state['errors'])) {
      foreach ($this->state['errors'] as $error) {
        $errors[] = [
          'message' => $error[0],
          'key' => $error[1],
        ];
      }
    }
    return $errors;
  }

  /**
   * Removes errors by key.
   */
  public function removeErrorsByKey(string $error_key): void {
    if (empty($this->state['errors'])) {
      return;
    }

    $this->state['errors'] = array_filter($this->state['errors'], function ($error) use ($error_key) {
      return $error[1] !== $error_key;
    });
  }

  /**
   * Returns state (full or partial).
   */
  public function getState(?string $key = NULL): mixed {
    return empty($key) ? $this->state : ($this->state[$key] ?? NULL);
  }

  /**
   * Returns stored errors, with or without debug data.
   */
  public function getErrorsTexts(?bool $unique = TRUE): array {
    if (!$this->hasErrors()) {
      return [];
    }

    $texts = [];
    foreach ($this->state['errors'] as $error) {
      if (!$unique || !in_array($error[0], $texts)) {
        $texts[] = $error[0];
      }
    }
    return $texts;
  }

  /**
   * Returns error keys.
   */
  public function getErrorsKeys(?bool $unique = TRUE): array {
    if (!$this->hasErrors()) {
      return [];
    }

    $keys = [];
    foreach ($this->state['errors'] as $error) {
      if (!$unique || !in_array($error[1], $keys)) {
        $keys[] = $error[1];
      }
    }
    return $keys;
  }

  /**
   * Returns last error text.
   */
  public function getLastErrorText(): string {
    return $this->getLastError()[0] ?? '';
  }

  /**
   * Returns last error text.
   */
  public function getLastError(): ?array {
    if (!$this->hasErrors()) {
      return NULL;
    }
    return end($this->state['errors']);
  }

}
